# -*- coding: utf-8 -*-

"""
This is the main module containing the class to be imported and used:
from aioharmony.client import HarmonyClient

The HarmonyClient class represents the Harmony Hub. Using the methods of
this class allows one to query or send commands to the Hub.
"""

import asyncio
import copy
import logging
from datetime import timedelta
from typing import List, NamedTuple, Optional, Union
from uuid import uuid4
from async_timeout import timeout

import aioharmony.exceptions as aioexc
import aioharmony.handler as handlers
from aioharmony.const import (
    CallbackType, ClientConfigType, HUB_COMMANDS, SendCommand,
    SendCommandDevice, SendCommandResponse
)
from aioharmony.helpers import call_callback, search_dict
from aioharmony.hubconnector import HubConnector, ConnectorCallbackType
from aioharmony.responsehandler import Handler, ResponseHandler

_LOGGER = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 60

ClientCallbackType = NamedTuple('ClientCallbackType',
                                [('connect', Optional[CallbackType]),
                                 ('disconnect', Optional[CallbackType]),
                                 ('new_activity', Optional[CallbackType]),
                                 ('config_updated', Optional[CallbackType])
                                 ])

# TODO: Add docstyle comments
# TODO: Clean up code styling


# pylint: disable=too-many-instance-attributes
class HarmonyClient:
    """An websocket client for connecting to the Logitech Harmony devices."""
    # pylint: disable=too-many-arguments
    def __init__(self,
                 ip_address: str,
                 callbacks: ClientCallbackType = None,
                 loop: asyncio.AbstractEventLoop = None) -> None:
        _LOGGER.debug("%s: Initialize HUB", ip_address)
        self._ip_address = ip_address
        self._callbacks = callbacks if callbacks is not None else \
            ClientCallbackType(None, None, None, None)
        self._loop = loop if loop else asyncio.get_event_loop()

        self._hub_config = ClientConfigType({}, {}, None, [], [])
        self._current_activity_id = None

        # Get the queue on which JSON responses will be put
        self._response_queue = asyncio.Queue()
        # Get the Hub Connection
        self._hub_connection = HubConnector(
            ip_address=self._ip_address,
            callbacks=ConnectorCallbackType(
                None,
                self._callbacks.disconnect
            ),
            response_queue=self._response_queue)
        # Get the Response Handler
        self._callback_handler = ResponseHandler(
            message_queue=self._response_queue)

        # Create the lock for sending commands or starting an activity
        self._snd_cmd_act_lck = asyncio.Lock()

        # Create the lock for getting HUB information.
        self._sync_lck = asyncio.Lock()

        # Create the activity start handler object
        handler = copy.copy(handlers.HANDLER_START_ACTIVITY_FINISHED)
        handler.handler_obj = self._update_activity_callback
        self._callback_handler.register_handler(
            handler=handler)

        # Create the notification handler object
        handler = copy.copy(handlers.HANDLER_NOTIFY)
        handler.handler_obj = self._notification_callback
        self._callback_handler.register_handler(
            handler=handler)

    @property
    def ip_address(self) -> str:
        return self._ip_address

    @property
    def name(self) -> Optional[str]:
        name = self._hub_config.info.get('friendlyName')
        return name if name is not None else self._ip_address

    @property
    def hub_config(self) -> ClientConfigType:
        return self._hub_config

    @property
    def callbacks(self) -> ClientCallbackType:
        return self._callbacks

    @callbacks.setter
    def callbacks(self, value: ClientCallbackType) -> None:
        self._callbacks = value
        self._hub_connection.callbacks = ConnectorCallbackType(
            connect=self._callbacks.connect,
            disconnect=self._callbacks.disconnect
        )

    @property
    def current_activity_id(self):
        return self._current_activity_id

    async def connect(self) -> bool:
        """

        :return: True if connection was successful, False if it was not.
        :rtype: bool
        :raises: :class:`~aioharmony.exceptions.TimeOut`
        """
        try:
            with timeout(DEFAULT_TIMEOUT):
                if not await self._hub_connection.connect():
                    return False
        except asyncio.TimeoutError:
            raise aioexc.TimeOut

        # Initiate a sync. That will then result in our notification handler
        # to receive the response and set our current config version
        # accordingly.

        results = await asyncio.gather(
            self.send_to_hub(command='get_current_state'),
            self.refresh_info_from_hub(),
            return_exceptions=True
        )
        for idx, result in enumerate(results):
            if isinstance(result, Exception):
                if not isinstance(
                        result,
                        aioexc.TimeOut):
                    raise result

                if idx == 0:
                    _LOGGER.error("%s: Timeout trying to sync hub.", self.name)

                continue

            if idx == 0:
                resp_data = result.get('data')
                if resp_data is not None:
                    self._hub_config = self._hub_config._replace(
                        config_version=resp_data.get('configVersion'))
                    _LOGGER.debug("%s: HUB configuration version is: %s",
                                  self.name,
                                  self._hub_config.config_version)

        if self._hub_connection.callbacks.connect is None and \
                self._callbacks.connect is not None:
            # First time call, add the callback handler now and run it.
            _LOGGER.debug("%s, calling connect callback for first time",
                          self.name)
            call_callback(
                callback_handler=self._callbacks.connect,
                result=self._ip_address,
                callback_uuid=self._ip_address,
                callback_name='connected'
            )
            self._hub_connection.callbacks = ConnectorCallbackType(
                self._callbacks.connect,
                self._callbacks.disconnect
            )
        return True

    async def close(self) -> None:
        """Close all connections and tasks

           This should be called to ensure everything is stopped and
           cancelled out.
        """
        if self._hub_connection:
            try:
                with timeout(DEFAULT_TIMEOUT):
                    await self._hub_connection.close()
            except asyncio.TimeoutError:
                raise aioexc.TimeOut

        if self._callback_handler:
            try:
                with timeout(DEFAULT_TIMEOUT):
                    await self._callback_handler.close()
            except asyncio.TimeoutError:
                raise aioexc.TimeOut

    async def disconnect(self) -> None:
        """Disconnect from Hub"""
        _LOGGER.debug("%s: Disconnecting from %s",
                      self.name,
                      self._ip_address)
        try:
            with timeout(DEFAULT_TIMEOUT):
                await self._hub_connection.disconnect()
        except asyncio.TimeoutError:
            raise aioexc.TimeOut

    async def refresh_info_from_hub(self) -> None:
        _LOGGER.debug("%s: Retrieving HUB information",
                      self.name)

        async with self._sync_lck:
            try:
                # Retrieve configuration and HUB version config.
                with timeout(DEFAULT_TIMEOUT):
                    results = await asyncio.gather(
                        self._get_config(),
                        self._retrieve_hub_info(),
                        return_exceptions=True
                    )
            except asyncio.TimeoutError:
                raise aioexc.TimeOut
            for idx, result in enumerate(results):
                if isinstance(
                        result,
                        aioexc.TimeOut):
                    # Timeout exception, just put out error then.
                    if idx == 0:
                        result_name = 'config'
                    elif idx == 1:
                        result_name = 'hub_info'
                    else:
                        result_name = 'get_current_activity'

                    _LOGGER.error("%s: Timeout trying to retrieve %s.",
                                  self.name,
                                  result_name)
                    return
                elif isinstance(result, Exception):
                    # Other exception, raise it.
                    raise result

            try:
                # Retrieve current activity, done only once config received.
                with timeout(DEFAULT_TIMEOUT):
                    await self._get_current_activity()
            except asyncio.TimeoutError:
                _LOGGER.error("%s: Timeout trying to retrieve current "
                              "activity.",
                              self.name)
                return

        # If we were provided a callback handler then call it now.
        if self._callbacks.config_updated:
            call_callback(
                callback_handler=self._callbacks.config_updated,
                result=self._hub_config.config,
                callback_uuid=self._ip_address,
                callback_name='config_updated_callback'
            )

    async def _get_config(self) -> Optional[dict]:
        """Retrieves the Harmony device configuration.

        Returns:
            A nested dictionary containing activities, devices, etc.
        """
        _LOGGER.debug("%s: Getting configuration",
                      self.name)
        # Send the command to the HUB
        response = await self.send_to_hub(command='get_config')
        if not response:
            # There was an issue
            return None

        if response.get('code') != 200:
            _LOGGER.error("%s: Incorrect status code %s received trying to "
                          "get configuration for %s",
                          self.name,
                          response.get('code'),
                          self._ip_address)
            return None

        self._hub_config = self._hub_config._replace(
            config=response.get('data'))

        self._hub_config = self._hub_config._replace(
            activities=list(
                {
                    'name': a['label'],
                    'name_lowercase': a['label'].lower(),
                    'id': int(a['id'])
                } for a in self._hub_config.config.get('activity')))

        self._hub_config = self._hub_config._replace(
            devices=list(
                {
                    'name': a['label'],
                    'name_lowercase': a['label'].lower(),
                    'id': int(a['id'])
                } for a in self._hub_config.config.get('device')))

        return self._hub_config.config

    async def _retrieve_hub_info(self) -> Optional[dict]:
        """Retrieve some information from the Hub."""
        try:
            with timeout(DEFAULT_TIMEOUT):
                response = await self._hub_connection.retrieve_hub_info()
        except asyncio.TimeoutError:
            raise aioexc.TimeOut

        if response is not None:
            self._hub_config = self._hub_config._replace(
                info=response)
        return self._hub_config.info

    async def send_to_hub(self,
                          command: str,
                          params: dict = None,
                          msgid: str = None,
                          wait: bool = True) -> Union[dict, bool]:

        if msgid is None:
            msgid = str(uuid4())

        if params is None:
            params = {
                'verb': 'get',
                'format': 'json'
            }

        response = None
        if wait:
            response = self._loop.create_future()
            resp_handler = Handler(handler_obj=response,
                                   handler_name=command,
                                   once=True,
                                   expiration=timedelta(
                                       seconds=DEFAULT_TIMEOUT)
                                   )
            self.register_handler(handler=resp_handler,
                                  msgid=msgid)

        try:
            with timeout(DEFAULT_TIMEOUT):
                if await self._hub_connection.send(
                        command='{}?{}'.format(
                            HUB_COMMANDS[command]['mime'],
                            HUB_COMMANDS[command]['command']
                        ),
                        params=params,
                        msgid=msgid
                ) is None:
                    # There was an issue
                    return False
        except asyncio.TimeoutError:
            raise aioexc.TimeOut

        if not wait:
            return True

        # Wait for the response to be available.
        try:
            with timeout(DEFAULT_TIMEOUT):
                await response
        except asyncio.TimeoutError:
            raise aioexc.TimeOut

        return response.result()

    async def _get_current_activity(self) -> bool:
        """Update current activity when changed."""
        _LOGGER.debug("%s: Retrieving current activity", self.name)

        # Send the command to the HUB
        try:
            response = await self.send_to_hub(command='get_current_activity')
        except aioexc.TimeOut:
            _LOGGER.error("%s: Timeout trying to retrieve current activity.",
                          self.name)
            response = None

        if not response:
            # There was an issue
            return False

        if response.get('code') != 200:
            _LOGGER.error("%s: Incorrect status code %s received trying to get"
                          "current activity for %s",
                          self.name,
                          response.get('code'),
                          self._ip_address)
            return False

        self._current_activity_id = int(response['data']['result'])
        _LOGGER.debug("%s: Current activity: %s(%s)",
                      self.name,
                      self.get_activity_name(self._current_activity_id),
                      self._current_activity_id)

        # If we were provided a callback handler then call it now.
        if self._callbacks.new_activity:
            call_callback(
                callback_handler=self._callbacks.new_activity,
                result=(self._current_activity_id,
                        self.get_activity_name(
                            activity_id=self._current_activity_id)
                        ),
                callback_uuid=self._ip_address,
                callback_name='new_activity_callback'
            )
        return True

    # pylint: disable=broad-except
    async def _notification_callback(self,
                                     message: dict = None) -> None:
        # We received a notification, check if the config version has changed.
        _LOGGER.debug("%s: Notification was received", self.name)
        resp_data = message.get('data')
        if resp_data is not None:
            current_hub_config_version = resp_data.get('configVersion')
            sync_status = resp_data.get('syncStatus')

            # If no sync status or it is 1 (sync in progress) then nothing
            # to do
            if current_hub_config_version is None or sync_status is None or \
                    sync_status == 1:
                return

            # Only do config update
            if current_hub_config_version != \
                    self._hub_config.config_version:
                _LOGGER.debug("%s: HUB configuration updated from version "
                              "%s to %s",
                              self.name,
                              self._hub_config.config_version,
                              current_hub_config_version)
                self._hub_config = self._hub_config._replace(
                    config_version=current_hub_config_version)
                # Get all the HUB information.
                await self.refresh_info_from_hub()

    # pylint: disable=broad-except
    async def _update_activity_callback(self,
                                        message: dict = None) -> None:
        """Update current activity when changed."""
        _LOGGER.debug("%s: New activity was started", self.name)

        new_activity = None
        message_data = message.get('data')
        if message_data is not None:
            new_activity = int(message_data.get('activityId'))

        if new_activity is None:
            await self._get_current_activity()
            return

        self._current_activity_id = new_activity
        _LOGGER.debug("%s: New activity: %s(%s)",
                      self.name,
                      self.get_activity_name(self._current_activity_id),
                      self._current_activity_id)

        # If we were provided a callback handler then call it now.
        if self._callbacks.new_activity:
            call_callback(
                callback_handler=self._callbacks.new_activity,
                result=(self._current_activity_id,
                        self.get_activity_name(
                            self._current_activity_id)
                        ),
                callback_uuid=self._ip_address,
                callback_name='new_activity_callback'
            )

    # pylint: disable=too-many-statements
    # pylint: disable=too-many-locals
    async def start_activity(self, activity_id) -> tuple:
        """Starts an activity.

        Args:
            activity_id: An int or string identifying the activity to start

        Returns:
            True if activity started, otherwise False
        """
        _LOGGER.debug("%s: Starting activity %s (%s)",
                      self.name,
                      self.get_activity_name(activity_id),
                      activity_id)
        params = {
            "async": "true",
            "timestamp": 0,
            "args": {
                "rule": "start"
            },
            "activityId": str(activity_id)
        }
        msgid = str(uuid4())

        activity_completed = self._loop.create_future()
        handler_list = []

        def register_activity_handler(activity_handler: Handler) -> None:
            handler_uuid = self.register_handler(
                handler=activity_handler,
                msgid=msgid
            )
            handler_list.append((handler_uuid, activity_handler))

        def unregister_handlers() -> None:
            _LOGGER.debug("%s: Unregistering handlers for Start Activity",
                          self.name)
            for activity_handler in handler_list:
                if not self.unregister_handler(activity_handler[0]):
                    _LOGGER.warning("%s: Callback %s with UUID %s was not "
                                    "found anymore",
                                    self.name,
                                    activity_handler[1].handler_name,
                                    activity_handler[0])

        def set_result(result: tuple) -> None:
            if activity_completed.done():
                _LOGGER.debug("%s: Result was already set through a previous "
                              "message.",
                              self.name)
                return
            activity_completed.set_result(result)

        def startactivity_started_callback(message: dict):
            if message.get('code') not in [100, 200]:
                _LOGGER.debug("%s: RunActivity code error: %s",
                              self.name,
                              message.get('code'))
                set_result((False,
                            message.get('msg')))

        def startactivity_in_progress_callback(message: dict) -> None:
            data = message.get('data')
            if data is not None:
                progress = {
                    'completed': data.get('done'),
                    'total': data.get('total')
                }
                _LOGGER.info("%s: %s/%s of start activity %s completed.",
                             self.name,
                             progress['completed'],
                             progress['total'],
                             activity_id)

        def startactivity_completed_callback(message: dict) -> None:
            if message.get('code') == 200:
                _LOGGER.debug("%s: Start or discrete completion code: %s",
                              self.name,
                              message.get('code'))
                set_result((True, message.get('msg')))
                return

            if message.get('code') != 100:
                _LOGGER.debug("%s: Start or discrete code error: %s",
                              self.name,
                              message.get('code'))
                set_result((False, message.get('msg')))
                return

        # Register handler to identify failure for initiating the start of
        # the activity
        handler = copy.copy(handlers.HANDLER_RUN_ACTIVITY)
        handler.handler_obj = startactivity_started_callback
        register_activity_handler(activity_handler=handler)

        # Register first handler to track progress of the start activity
        handler = copy.copy(handlers.HANDLER_START_ACTIVITY_IN_PROGRESS)
        handler.handler_obj = startactivity_in_progress_callback
        register_activity_handler(activity_handler=handler)

        # Register second handler to track progress of the start activity
        handler = copy.copy(handlers.HANDLER_HELPDISCRETES)
        handler.handler_obj = startactivity_in_progress_callback
        register_activity_handler(activity_handler=handler)

        handler = copy.copy(handlers.HANDLER_START_ACTIVITY_COMPLETE)
        handler.handler_obj = startactivity_completed_callback
        register_activity_handler(activity_handler=handler)

        # Get the lock ensuring we're the only ones able to initiate
        async with self._snd_cmd_act_lck:
            response = await self.send_to_hub(command='start_activity',
                                              params=params,
                                              msgid=msgid,
                                              wait=False)
            if not response:
                unregister_handlers()

            try:
                with timeout(DEFAULT_TIMEOUT):
                    status = await activity_completed
            except asyncio.TimeoutError:
                raise aioexc.TimeOut
            finally:
                unregister_handlers()
                _LOGGER.debug("%s: Start activity %s (%s) has been completed",
                              self.name,
                              self.get_activity_name(activity_id),
                              activity_id)

        return status

    async def send_commands(self,
                            commands: List[SendCommand]) -> \
            List[SendCommandResponse]:

        _LOGGER.debug("%s: Sending commands to HUB", self.name)
        # Get the lock ensuring we're the only ones able to initiate
        command_future_list = []
        msgid_dict = {}
        async with self._snd_cmd_act_lck:
            for command in commands:
                if isinstance(command, (float, int)):
                    await asyncio.sleep(command)
                    continue

                # Create the future to be set for the result.
                # The HUB sends a message back if there is an issue with
                # the command, otherwise it won't sent anything back.
                command_future_list.append(self._loop.create_future())

                expiration = 0.5
                if command.delay is not None:
                    expiration += command.delay

                command_handler = Handler(handler_obj=command_future_list[-1],
                                          handler_name='{}_{}'.format(
                                              command.device,
                                              command.command),
                                          expiration=timedelta(
                                              seconds=expiration)
                                          )

                msgid = await self._send_command(command,
                                                 command_handler)
                if msgid is not None:
                    msgid_dict.update({msgid: command})

        # Go through the result list to determine there were any issues with
        # any of the commands sent. Only if there is an issue would a response
        # have been received.
        done, _ = await asyncio.wait(command_future_list, timeout=1)

        error_response_list = []
        for result_returned in done:
            result = result_returned.result()
            msgid = result.get('id')
            if msgid is None:
                _LOGGER.warning("%s: Received response for send commands "
                                "without a message id",
                                self.name)
                continue

            command_sent = msgid_dict.get(msgid)
            if command_sent is None:
                _LOGGER.warning("%s: Received response for send command "
                                "with unknown message id %s",
                                self.name,
                                msgid)
                continue

            if isinstance(command_sent, (float, int)):
                continue

            _LOGGER.debug("%s: Received code %s for command %s to device "
                          "%s: %s",
                          self.name,
                          result.get('code'),
                          command_sent.command,
                          command_sent.device,
                          result.get('msg')
                          )

            # HUB might send back OK (200) code, ignore those.
            if str(result.get('code')) != '200':
                error_response_list.append(SendCommandResponse(
                    command=command_sent,
                    code=result.get('code'),
                    msg=result.get('msg')
                ))

        _LOGGER.debug("%s: Sending commands to HUB has been completed",
                      self.name)
        return error_response_list

    async def _send_command(self,
                            command: SendCommandDevice,
                            callback_handler: Handler) -> Optional[str]:
        """Send a command to specified device

        :param command: Command to send to the device. (device, command, delay)
        :type command: SendCommandDevice
        :return: msgid with which this command was sent for.
        :rtype: str

        """
        _LOGGER.debug("%s: Sending command %s to device %s (%s) with delay "
                      "%ss",
                      self.name,
                      command.command,
                      self.get_device_name(command.device),
                      command.device,
                      command.delay)
        params = {
            "status": "press",
            "timestamp": '0',
            "verb": "render",
            "action": '{{"command": "{}",'
                      '"type": "IRCommand",'
                      '"deviceId": "{}"}}'.format(command.command,
                                                  command.device)
        }
        msgid = str(uuid4())

        # Register the handler for this command.
        self.register_handler(handler=callback_handler,
                              msgid=msgid)

        # Send the command to the HUB
        response = await self.send_to_hub(command='send_command',
                                          params=params,
                                          msgid=msgid,
                                          wait=False)
        if not response:
            # There was an issue
            return None

        if command.delay > 0:
            await asyncio.sleep(command.delay)

        params['status'] = 'release'
        # Send the command to the HUB
        await self.send_to_hub(command='send_command',
                               params=params,
                               msgid=msgid,
                               wait=False)

        _LOGGER.debug("%s: Sending command %s to device %s (%s) with delay "
                      "%ss has been completed",
                      self.name,
                      command.command,
                      self.get_device_name(command.device),
                      command.device,
                      command.delay)
        return msgid

    def get_activity_id(self, activity_name) -> Optional[int]:
        """Find the activity ID for the provided activity name."""
        if activity_name is None:
            return None

        item = search_dict(match_value=activity_name.lower(),
                           key='name_lowercase',
                           search_list=self._hub_config.activities)
        return item.get('id') if item else None

    def get_activity_name(self, activity_id) -> Optional[str]:
        """Find the activity name for the provided ID."""
        if activity_id is None:
            return None

        item = search_dict(match_value=int(activity_id),
                           key='id',
                           search_list=self._hub_config.activities)
        return item.get('name') if item else None

    def get_device_id(self, device_name) -> Optional[int]:
        """Find the device ID for the provided device name."""
        if device_name is None:
            return None

        item = search_dict(match_value=device_name.lower(),
                           key='name_lowercase',
                           search_list=self._hub_config.devices)
        return item.get('id') if item else None

    def get_device_name(self, device_id) -> Optional[str]:
        """Find the device name for the provided ID."""
        if device_id is None:
            return None

        item = search_dict(match_value=int(device_id),
                           key='id',
                           search_list=self._hub_config.devices)
        return item.get('name') if item else None

    def register_handler(self, *args, **kwargs) -> str:
        """
        Exposes
        :meth:`aioharmony.responsehandler.ResponseHandler.register_handler` for
        use to register other callbacks.

        See
        :meth:`~aioharmony.responsehandler.ResponseHandler.register_handler`
        in class
        :class:`~aioharmony.responsehandler.ResponseHandler` for further
        information on this method
        """
        return self._callback_handler.register_handler(*args, **kwargs)

    def unregister_handler(self, *args, **kwargs) -> bool:
        """
        Exposes
        :meth:`aioharmony.responsehandler.ResponseHandler.unregister_handler`
        for use to unregister callbacks.

        See
        :meth:`~aioharmony.responsehandler.ResponseHandler.unregister_handler`
        in class
        :class:`~aioharmony.responsehandler.ResponseHandler` for further
        information on this method
        """
        return self._callback_handler.unregister_handler(*args, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Module for querying and controlling Logitech Harmony devices."""

import argparse
import asyncio
import json
import logging
import sys
from typing import Optional

import aioharmony.exceptions
from aioharmony.harmonyapi import HarmonyAPI, SendCommandDevice
from aioharmony.responsehandler import Handler

# TODO: Add docstyle comments
# TODO: Clean up code styling

hub_client = None


async def get_client(ip_address, show_responses) -> Optional[HarmonyAPI]:
    client = HarmonyAPI(ip_address)

    def output_response(message):
        print(message)

    listen_callback = Handler(handler_obj=output_response,
                              handler_name='output_response',
                              once=False
                              )
    if show_responses:
        client.register_handler(handler=listen_callback)

    if await client.connect():
        print("Connected to HUB {} with firmware version {}".format(
            client.name,
            client.fw_version))
        return client

    print("An issue occured trying to connect")

    return None


async def just_listen(client, _):
    # Create handler to output everything.
    print("Starting to listen on HUB {} with firmware version {}".format(
        client.name,
        client.fw_version))


# Functions for use on command line
async def show_config(client, _):
    """Connects to the Harmony and return current configuration.
    """
    config = client.config

    if config:
        print(json.dumps(client.json_config, sort_keys=True, indent=4))
    else:
        print("There was a problem retrieving the configuration")


async def show_detailed_config(client, _):
    """Connects to the Harmony and return current configuration.
    """
    config = client.config

    if config:
        print(json.dumps(config, sort_keys=True, indent=4,
                         separators=(',', ': ')))
    else:
        print("There was a problem retrieving the configuration")


async def show_current_activity(client, _):
    """Returns Harmony hub's current activity.
    """
    activity_id, activity_name = client.current_activity

    if activity_name:
        print("{} ({})".format(activity_name, activity_id))
    elif activity_id:
        print(activity_id)
    else:
        print('Unable to retrieve current activity')


async def start_activity(client, args):
    """Connects to Harmony Hub and starts an activity

    Args:
        args (argparse): Argparse object containing required variables from
        command line

    """
    if args.activity is None:
        print("No activity provided to start")
        return

    if (args.activity.isdigit()) or (args.activity == '-1'):
        activity_id = args.activity
    else:
        activity_id = client.get_activity_id(args.activity)
        if activity_id:
            print('Found activity named %s (id %s)' % (args.activity,
                                                       activity_id))

    if activity_id:
        status = await client.start_activity(activity_id)

        if status[0]:
            print('Started Actvivity, message: ', status[1])
        else:
            print('Activity start failed: ', status[1])


async def power_off(client, _):
    """Power off Harmony Hub.
    """
    status = await client.power_off()

    if status:
        print('Powered Off')
    else:
        print('Power off failed')


async def send_command(client, args):
    """Connects to the Harmony and send a simple command.

    Args:
        args (argparse): Argparse object containing required variables from
        command line

    """
    device_id = None
    if args.device_id.isdigit():
        if client.get_device_name(int(args.device_id)):
            device_id = args.device_id

    if device_id is None:
        device_id = client.get_device_id(str(args.device_id).strip())

    if device_id is None:
        print("Device {} is invalid.".format(args.device_id))
        return

    snd_cmmnd = SendCommandDevice(
        device=device_id,
        command=args.command,
        delay=args.hold_secs)

    snd_cmmnd_list = []
    for _ in range(args.repeat_num):
        snd_cmmnd_list.append(snd_cmmnd)
        if args.delay_secs > 0:
            snd_cmmnd_list.append(args.delay_secs)

    result_list = await client.send_commands(snd_cmmnd_list)

    if result_list:
        for result in result_list:
            print("Sending of command {} to device {} failed with code {}: "
                  "{}".format(
                      result.command.command,
                      result.command.device,
                      result.code,
                      result.msg))
    else:
        print('Command Sent')


async def change_channel(client, args):
    """Change channel

    Args:
        args (argparse): Argparse object containing required variables from
        command line

    """
    status = await client.change_channel(args.channel)

    if status:
        print('Changed to channel {}'.format(args.channel))
    else:
        print('Change to channel {} failed'.format(args.channel))


# def discover(args):
#     hubs = harmony_discovery.discover()
#     pprint(hubs)


async def sync(client, _):
    """Syncs Harmony hub to web service.
    Args:
        args (argparse): Argparse object containing required variables from
        command line

    Returns:
        Completion status
    """
    status = await client.sync()

    if status:
        print('Sync complete')
    else:
        print("Sync failed")


async def run():
    """Main method for the script."""
    global hub_client

    parser = argparse.ArgumentParser(
        description='aioharmony - Harmony device control',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    required_flags = parser.add_mutually_exclusive_group(required=True)

    # Required flags go here.
    required_flags.add_argument('--harmony_ip',
                                help='IP Address of the Harmony device.')
    required_flags.add_argument('--discover',
                                action='store_true',
                                help='Scan for Harmony devices.')

    # Flags with default values go here.
    loglevels = dict((logging.getLevelName(level), level)
                     for level in [10, 20, 30, 40, 50])
    parser.add_argument('--harmony_port',
                        required=False,
                        default=5222,
                        type=int,
                        help=('Network port that the Harmony is listening'
                              ' on.'))
    parser.add_argument('--loglevel',
                        default='INFO',
                        choices=list(loglevels.keys()),
                        help='Logging level for all components to '
                             'print to the console.')

    show_responses_parser = parser.add_mutually_exclusive_group(
        required=False)
    show_responses_parser.add_argument('--show_responses',
                                       dest='show_responses',
                                       action='store_true',
                                       help='Print out responses coming from '
                                            'HUB.')
    show_responses_parser.add_argument('--no-show_responses',
                                       dest='show_responses',
                                       action='store_false',
                                       help='Do not print responses coming '
                                            'from HUB.')
    show_responses_parser.set_defaults(show_responses=False)

    parser.add_argument('--wait',
                        required=False,
                        default=0,
                        type=int,
                        help='How long to wait in seconds after completion, '
                             'useful in combination with --show-responses.\n'
                             'Use -1 to wait infinite, otherwise has to be a '
                             'positive number.')

    subparsers = parser.add_subparsers()

    show_config_parser = subparsers.add_parser(
        'show_config', help='Print the Harmony device configuration.')
    show_config_parser.set_defaults(func=show_config)

    show_detailed_config_parser = subparsers.add_parser(
        'show_detailed_config', help='Print the detailed Harmony device'
                                     ' configuration.')
    show_detailed_config_parser.set_defaults(func=show_detailed_config)

    show_activity_parser = subparsers.add_parser(
        'show_current_activity', help='Print the current activity config.')
    show_activity_parser.set_defaults(func=show_current_activity)

    start_activity_parser = subparsers.add_parser(
        'start_activity', help='Switch to a different activity.')
    start_activity_parser.add_argument(
        '--activity', help='Activity to switch to, id or label.')
    start_activity_parser.set_defaults(func=start_activity)

    power_off_parser = subparsers.add_parser(
        'power_off', help='Stop the activity.')
    power_off_parser.set_defaults(func=power_off)

    sync_parser = subparsers.add_parser('sync', help='Sync the harmony.')
    sync_parser.set_defaults(func=sync)

    listen_parser = subparsers.add_parser('listen', help='Output everything '
                                                         'HUB sends out')
    listen_parser.set_defaults(func=just_listen)

    command_parser = subparsers.add_parser(
        'send_command', help='Send a simple command.')
    command_parser.add_argument(
        '--device_id',
        help='Specify the device id to which we will send the command.')
    command_parser.add_argument(
        '--command', help='IR Command to send to the device.')
    command_parser.add_argument(
        '--repeat_num', type=int, default=1,
        help='Number of times to repeat the command. Defaults to 1')
    command_parser.add_argument(
        '--delay_secs', type=float, default=0.4,
        help='Delay between sending repeated commands. Not used if only '
             'sending a single command. Defaults to 0.4 seconds')
    command_parser.add_argument(
        '--hold_secs', type=float, default=0,
        help='Number of seconds to "hold" before releasing. Defaults to 0.4'
             'seconds')
    command_parser.set_defaults(func=send_command)

    change_channel_parser = subparsers.add_parser(
        'change_channel', help='Change the channel')
    change_channel_parser.add_argument(
        '--channel', help='Channel to switch to.')
    change_channel_parser.set_defaults(func=change_channel)

    args = parser.parse_args()

    logging.basicConfig(
        level=loglevels[args.loglevel],
        format='%(asctime)s:%(levelname)s:\t%(name)s\t%(message)s')

    if args.wait < 0 and args.wait != -1:
        print("Invalid value provided for --wait.")
        parser.print_help()
        return

    if args.discover:
        # discover(args)
        pass
    else:
        coroutine = None
        if not hasattr(args, 'func') and not args.show_responses:
            parser.print_help()
            return

        # Connect to the HUB
        try:
            hub_client = await get_client(args.harmony_ip,
                                          args.show_responses)
            if hub_client is None:
                return
        except aioharmony.exceptions.TimeOut:
            print("Action did not complete within a reasonable time.")
            raise

        if hasattr(args, 'func'):
            coroutine = args.func(hub_client, args)

        # Execute provided request.
        try:
            if coroutine is not None:
                await coroutine
        except aioharmony.exceptions.TimeOut:
            print("Action did not complete within a reasonable time.")
            raise

        # Now sleep for provided time.
        if args.wait >= 0:
            await asyncio.sleep(args.wait)
        else:
            while True:
                await asyncio.sleep(60)

        if hub_client:
            await asyncio.wait_for(hub_client.close(), timeout=10)
            hub_client = None


def main() -> None:
    loop = asyncio.new_event_loop()
    try:
        loop.run_until_complete(run())
        while asyncio.all_tasks(loop):
            loop.run_until_complete(asyncio.gather(*asyncio.all_tasks(loop)))
        loop.close()
    except KeyboardInterrupt:
        print("Exit requested.")
        if hub_client is not None:
            loop.run_until_complete(
                asyncio.wait_for(hub_client.close(), timeout=10)
            )
        print("Closed.")


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-

"""
This is module specifies the Handler class.

A number of HANDLER constants have been defined here, do not USE these,
copy them instead.

"""
import re
from datetime import timedelta
from aioharmony.const import CallbackType

DEFAULT_TIMEOUT = 60


class Handler:
    """
    This class is to be used to create Handler objects for registering new
    callbacks through method
    :meth:`aioharmony.client.HarmonyClient.register_handler`

    **handler_obj:** This parameter can accept one of the following types of
    objects.


    * Future : result of the future will be set to the json that was received.

    * Event : event will be triggered (set). Event will not be cleared, this
      is the responsibility of the object listening for the event.

    * Coroutine : Coroutine will be put on the task loop with one (1)
      argument containing the json response.

    * Callback  : Callback will be called with one (1) argument containing
      the json response.

    :param handler_obj: Object to be called or set.
    :type handler_obj: CallbackType
    :param handler_name: Descriptive name for the handler. \
                         DEFAULT = None
    :type handler_name: Optional[str]
    :param resp_json: A JSON dict used to match the response upon.
                      DEFAULT = None
    :type resp_json: dict
    :param once: Only execute callback once (True) or keep callback until
                removed explicitly (False)
                DEFAULT = True
    :type once: Optional[bool]
    :param expiration: How long handler should be active once registered
                       before being removed.
                       DEFAULT = None
    :type expiration: Optional[timedelta]

    """

    # pylint: disable=too-many-arguments
    def __init__(self,
                 handler_obj: CallbackType,
                 handler_name: str = None,
                 resp_json: dict = None,
                 once: bool = True,
                 expiration: timedelta = None) -> None:
        self._handler_obj = handler_obj
        self._handler_name = handler_name
        self._resp_json = resp_json
        self._once = once
        self._expiration = expiration

    def __copy__(self):
        json_resp = dict()
        json_resp.update(self._resp_json)
        new_handler = Handler(
            handler_obj=self._handler_obj,
            handler_name=self._handler_name,
            resp_json=json_resp,
            once=self._once,
            expiration=self._expiration
        )
        return new_handler

    @property
    def handler_obj(self) -> CallbackType:
        """

        :param value: New handler_obj
        :type value: CallbackType
        :return: Returns handler_obj
        :rtype: CallbackType
        """
        return self._handler_obj

    @handler_obj.setter
    def handler_obj(self, value: CallbackType) -> None:
        """"""
        self._handler_obj = value

    @property
    def handler_name(self) -> str:
        """

        :param value: New handler_name
        :type value: str
        :return: Returns handler_name
        :rtype: str
        """
        return self._handler_name

    @handler_name.setter
    def handler_name(self, value: str) -> None:
        """"""
        self._handler_name = value

    @property
    def resp_json(self) -> dict:
        """

        :param value: New resp_type
        :type value: dict
        :return: Returns resp_type
        :rtype: dict
        """
        return self._resp_json

    @resp_json.setter
    def resp_json(self, value: dict) -> None:
        """"""
        self._resp_json = value

    @property
    def once(self) -> bool:
        """

        :param value: New once
        :type value: bool
        :return: Returns once
        :rtype: bool
        """
        return self._once

    @once.setter
    def once(self, value: bool) -> None:
        """"""
        self._once = value

    @property
    def expiration(self) -> timedelta:
        """

        :param value: New expiration
        :type value: timedelta
        :return: Returns expiration
        :rtype: timedelta
        """
        return self._expiration

    @expiration.setter
    def expiration(self, value: timedelta) -> None:
        """"""
        self._expiration = value


#
#
# Define a number of common handlers.
#
# These are NOT to be used as such, instead they HAVE TO BE copied.
#
#


def dummy_callback(message):
    return message


HANDLER_NOTIFY = Handler(
    handler_obj=dummy_callback,
    handler_name='Notification_Received',
    resp_json={
        'type': re.compile(r'^connect\.stateDigest\?notify$')
    },
    once=False
)

HANDLER_START_ACTIVITY_FINISHED = Handler(
    handler_obj=dummy_callback,
    handler_name='Activity_Changed',
    resp_json={
        'type': re.compile(r'^harmony\.engine\?startActivityFinished$')
    },
    once=False
)

HANDLER_RUN_ACTIVITY = Handler(
    handler_obj=dummy_callback,
    handler_name='runactivity',
    resp_json={
        'cmd': re.compile(r'^harmony\.activityengine\?runactivity$')
    },
    once=False,
    expiration=timedelta(seconds=DEFAULT_TIMEOUT * 5)
)

HANDLER_START_ACTIVITY_IN_PROGRESS = Handler(
    handler_obj=dummy_callback,
    handler_name='progress_startactivity',
    resp_json={
        'cmd':  re.compile(r'^harmony\.engine\?startActivity$'),
        'code': 100
    },
    once=False,
    expiration=timedelta(seconds=DEFAULT_TIMEOUT * 5)
)

HANDLER_HELPDISCRETES = Handler(
    handler_obj=dummy_callback,
    handler_name='progress_discrete',
    resp_json={
        'cmd':  re.compile(r'^harmony\.engine\?helpdiscretes$'),
        'data': {
            'isHelpDiscretes': 'true'
        }
    },
    once=False,
    expiration=timedelta(seconds=DEFAULT_TIMEOUT * 5)
)

HANDLER_START_ACTIVITY_COMPLETE = Handler(
    handler_obj=dummy_callback,
    handler_name='startactivity_or_discrete',
    resp_json={
        'cmd': re.compile(
            r'(^harmony\.engine\?helpdiscretes$)|'
            r'(^harmony\.engine\?startActivity$)'
        )
    },
    once=False,
    expiration=timedelta(seconds=DEFAULT_TIMEOUT * 5)
)



# -*- coding: utf-8 -*-

"""Connector class for connecting to send requests and receive

responses."""

import asyncio
import logging
import socket
from contextlib import suppress
from typing import NamedTuple, Optional
from urllib.parse import urlparse
from uuid import uuid4

import aiohttp
from async_timeout import timeout

from aioharmony.handler import CallbackType
from aioharmony.helpers import call_callback

DEFAULT_DOMAIN = 'svcs.myharmony.com'
DEFAULT_HUB_PORT = '8088'
DEFAULT_TIMEOUT = 30

_LOGGER = logging.getLogger(__name__)


# TODO: Add docstyle comments
# TODO: Clean up code styling

ConnectorCallbackType = NamedTuple('ConnectorCallbackType',
                                   [('connect', Optional[CallbackType]),
                                    ('disconnect', Optional[CallbackType])
                                    ])


# pylint: disable=too-many-instance-attributes
class HubConnector:
    """An websocket client for connecting to the Logitech Harmony devices."""

    def __init__(self,
                 ip_address: str,
                 response_queue: asyncio.Queue,
                 callbacks: ConnectorCallbackType = None,
                 auto_reconnect=True) -> None:
        self._ip_address = ip_address
        self._response_queue = response_queue
        self._callbacks = callbacks if callbacks is not None else \
            ConnectorCallbackType(None, None)
        self._auto_reconnect = auto_reconnect

        self._remote_id = None
        self._domain = DEFAULT_DOMAIN

        self._aiohttp_session = None
        self._websocket = None

        self._connect_disconnect_lock = asyncio.Lock()

        self._listener_task = None

        self._connected = False

    @property
    def callbacks(self) -> ConnectorCallbackType:
        """Return callbacks."""
        return self._callbacks

    @callbacks.setter
    def callbacks(self, value: ConnectorCallbackType) -> None:
        """Set callbacks."""
        self._callbacks = value

    async def close(self):
        """Close all connections and tasks

           This should be called to ensure everything is stopped and
           cancelled out.
        """
        # Close connections.
        await self.disconnect()

    @property
    def _session(self):
        """Create the aiohttp client session if not existing."""
        # Set connection timeout. Default total timeout is 5 minutes.
        if self._aiohttp_session:
            return self._aiohttp_session

        # Specify socket
        conn = aiohttp.TCPConnector(
            family=socket.AF_INET,
            verify_ssl=False,
        )

        session_timeout = aiohttp.ClientTimeout(connect=DEFAULT_TIMEOUT)
        self._aiohttp_session = aiohttp.ClientSession(connector=conn,
                                                      timeout=session_timeout)
        return self._aiohttp_session

    async def _get_remote_id(self) -> Optional[str]:
        """Retrieve remote id from the HUB."""

        if self._remote_id is None:
            # We do not have the remoteId yet, get it first.
            response = await self.retrieve_hub_info()
            if response is not None:
                self._remote_id = response.get('activeRemoteId')
                domain = urlparse(response.get('discoveryServer'))
                self._domain = domain.netloc if domain.netloc else \
                    DEFAULT_DOMAIN
        return self._remote_id

    async def connect(self, is_reconnect: bool = False) -> bool:
        """Connect to Hub Web Socket"""
        # Acquire the lock.
        if self._connect_disconnect_lock.locked():
            _LOGGER.debug("%s: Waiting for other connect", self._ip_address)

        async with self._connect_disconnect_lock:
            # Return connected if we are already connected.
            if self._websocket and not self._websocket.closed:
                return True

            _LOGGER.debug("%s: Starting connect.", self._ip_address)

            if is_reconnect:
                log_level = 10
            else:
                log_level = 40

            if await self._get_remote_id() is None:
                # No remote ID means no connect.
                _LOGGER.log(log_level,
                            "%s: Unable to retrieve HUB id",
                            self._ip_address)
                return False

            _LOGGER.debug("%s: Connecting for hub %s", self._ip_address,
                          self._remote_id)
            try:
                self._websocket = await self._session.ws_connect(
                    'ws://{}:{}/?domain={}&hubId={}'.format(
                        self._ip_address,
                        DEFAULT_HUB_PORT,
                        self._domain,
                        self._remote_id
                    ),
                    heartbeat=10
                )
            except (aiohttp.ServerTimeoutError, aiohttp.ClientError,
                    aiohttp.WSServerHandshakeError) as exc:
                if not is_reconnect:
                    if isinstance(exc, aiohttp.ServerTimeoutError):
                        _LOGGER.log(log_level,
                                    "%s: Connection timed out for hub %s",
                                    self._ip_address,
                                    self._remote_id)
                    elif isinstance(exc, aiohttp.ClientError):
                        _LOGGER.log(log_level,
                                    "%s: Exception trying to establish web "
                                    "socket connection for hub %s: %s",
                                    self._ip_address,
                                    self._remote_id,
                                    exc)
                    else:
                        _LOGGER.log(log_level,
                                    "%s: Invalid status code %s received "
                                    "trying to connect for hub %s: %s",
                                    self._ip_address,
                                    exc.status,
                                    self._remote_id,
                                    exc)
                self._websocket = None
                return False

            # Now put the listener on the loop.
            if not self._listener_task:
                self._listener_task = asyncio.ensure_future(
                    self._listener(self._websocket))

            # Set connected to True, disconnect sets this to False to
            # prevent automatic reconnect when disconnect is explicitly called
            self._connected = True
            call_callback(callback_handler=self._callbacks.connect,
                          result=self._ip_address,
                          callback_uuid=self._ip_address,
                          callback_name='connected'
                          )
            return True

    async def disconnect(self) -> None:
        """Disconnect from Hub"""
        _LOGGER.debug("%s: Disconnecting", self._ip_address)
        # Acquire the lock.
        async with self._connect_disconnect_lock:
            # Set connected to false preventing reconnect from trying to
            # reconnect.
            self._connected = False

            if self._websocket:
                with suppress(asyncio.TimeoutError), \
                     timeout(DEFAULT_TIMEOUT):
                    await self._websocket.close()

                await self._session.close()
                # Zero-sleep to allow underlying connections to close.
                await asyncio.sleep(0)

            # Stop our listener if still running
            if self._listener_task and not self._listener_task.done():
                self._listener_task.cancel()
                # One more wait ensuring that our tasks are cancelled.
                await asyncio.sleep(0)

            self._websocket = None

    async def _reconnect(self) -> None:
        """Perform reconnect to HUB if connection failed"""
        call_callback(callback_handler=self._callbacks.disconnect,
                      result=self._ip_address,
                      callback_uuid=self._ip_address,
                      callback_name='disconnected'
                      )
        if not self._connected:
            _LOGGER.debug("%s: Connection was closed through "
                          "disconnect, not reconnecting",
                          self._ip_address)
            return

        _LOGGER.debug("%s: Connection closed, reconnecting",
                      self._ip_address)
        is_reconnect = False
        while not await self.connect(is_reconnect=is_reconnect):
            await asyncio.sleep(10)
            is_reconnect = True

    async def send(self, command, params, msgid=None) -> Optional[str]:
        """Send a payload request to Harmony Hub and return json response."""
        # Make sure we're connected.
        if not await self.connect():
            return

        if not msgid:
            msgid = str(uuid4())

        payload = {
            "hubId": self._remote_id,
            "timeout": DEFAULT_TIMEOUT,
            "hbus": {
                "cmd": command,
                "id": msgid,
                "params": params
            }
        }

        _LOGGER.debug("%s: Sending payload: %s", self._ip_address, payload)
        try:
            await self._websocket.send_json(payload)
        except aiohttp.ClientError as exc:
            _LOGGER.error("%s: Exception sending payload: %s",
                          self._ip_address, exc)
            return

        return msgid

    async def post(self, url, json_request, headers=None) -> Optional[dict]:
        """Post a json request and return the response."""
        _LOGGER.debug("%s: Sending post request: %s",
                      self._ip_address,
                      json_request)
        try:
            async with self._session.post(
                    url, json=json_request, headers=headers) as response:
                json_response = await response.json(content_type=None)
                _LOGGER.debug("%s: Post response: %s",
                              self._ip_address,
                              json_response)
        except aiohttp.ClientError as exc:
            _LOGGER.error("%s: Exception on post: %s", self._ip_address, exc)
        else:
            return json_response

        return None

    # pylint: disable=broad-except
    async def _listener(self, websocket=None) -> None:
        """Listen for  messages on web socket"""
        _LOGGER.debug("%s: Listener started", self._ip_address)
        if not websocket:
            websocket = self._websocket

        if not websocket:
            _LOGGER.error("%s: No web socket to listen on", self._ip_address)
        # This is a continuous loop until socket is closed.
        have_connection = True
        while have_connection:
            # Put everything here in a try block, we do not want this
            # to stop running out due to an exception.
            try:
                if not websocket or websocket.closed:
                    _LOGGER.debug("%s: Web socket has been closed")
                    have_connection = False
                    break

                try:
                    response = await websocket.receive()
                except aiohttp.ClientError as exc:
                    _LOGGER.error("%s: Exception during receive: %s",
                                  self._ip_address, exc)
                    break

                _LOGGER.debug("%s: Response payload: %s", self._ip_address,
                              response.data)

                if response.type == aiohttp.WSMsgType.CLOSED:
                    _LOGGER.debug("%s: Web socket closed",
                                  self._ip_address)
                    # Issue reconnect event if enabled
                    have_connection = False
                    break

                if response.type == aiohttp.WSMsgType.ERROR:
                    _LOGGER.error("%s: Response error", self._ip_address)
                    continue

                if response.type != aiohttp.WSMsgType.TEXT:
                    continue

                response_json = response.json()
                if not response_json:
                    continue

                # Put response on queue.
                self._response_queue.put_nowait(response_json)

            except asyncio.CancelledError:
                _LOGGER.debug("%s: Received STOP for listener",
                              self._ip_address)
                break

            # pylint: disable=broad-except
            # Need to catch everything here to prevent an issue in a
            # callback from ever causing the handler to exit.
            except Exception as exc:
                _LOGGER.exception("%s: Exception in listener: %s",
                                  self._ip_address, exc)

        self._listener_task = None
        _LOGGER.debug("%s: Listener stopped.", self._ip_address)

        # If we exited the loop due to connection closed then
        # call reconnect to deterimine if we should reconnect again.
        if not have_connection:
            await self._reconnect()

    async def retrieve_hub_info(self) -> Optional[dict]:
        """Retrieve the harmony Hub information."""
        _LOGGER.debug("%s: Retrieving Harmony Hub information.",
                      self._ip_address)
        url = 'http://{}:{}/'.format(self._ip_address, DEFAULT_HUB_PORT)
        headers = {
            'Origin': 'http://sl.dhg.myharmony.com',
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'Accept-Charset': 'utf-8',
        }
        json_request = {
            "id ": 1,
            "cmd": "setup.account?getProvisionInfo",
            "params": {}
        }

        response = await self.post(url, json_request, headers)

        if response is not None:
            return response.get('data')

        return None



# -*- coding: utf-8 -*-

"""
Constants used throughout the modules
"""

import asyncio
from typing import Any, Callable, List, NamedTuple, Optional, Union

#
# DEFAULT values
#
DEFAULT_CMD = 'vnd.logitech.connect'
DEFAULT_DISCOVER_STRING = '_logitech-reverse-bonjour._tcp.local.'
DEFAULT_HUB_PORT = '8088'
DEFAULT_HARMONY_MIME = 'vnd.logitech.harmony/vnd.logitech.harmony.engine'

#
# The HUB commands that can be send
#
HUB_COMMANDS = {
    'change_channel':       {
        'mime':    'harmony.engine',
        'command': 'changeChannel'
    },
    'get_current_state':    {
        'mime': 'vnd.logitech.connect/vnd.logitech.statedigest',
        'command': 'get'
    },
    'get_config':           {
        'mime':    DEFAULT_HARMONY_MIME,
        'command': 'config'
    },
    'get_current_activity': {
        'mime':    DEFAULT_HARMONY_MIME,
        'command': 'getCurrentActivity'
    },
    'send_command':         {
        'mime':    DEFAULT_HARMONY_MIME,
        'command': 'holdAction'
    },
    'start_activity':       {
        'mime':    'harmony.activityengine',
        'command': 'runactivity'
    },
    'sync':                 {
        'mime':    'setup.sync',
        'command': None
    }
}

#
# Different types used within aioharmony
#

# Type for callback parameters
CallbackType = Union[
    asyncio.Future,
    asyncio.Event,
    Callable[[object, Optional[Any]], Any]
]

ClientConfigType = NamedTuple('ClientConfigType',
                              [('config', dict),
                               ('info', dict),
                               ('config_version', Optional[int]),
                               ('activities', List[dict]),
                               ('devices', List[dict])
                               ])

# Type for a command to send to the HUB
SendCommandDevice = NamedTuple('SendCommandDevice',
                               [('device', int),
                                ('command', str),
                                ('delay', float)
                                ])

# Type for send command to aioharmony,
SendCommand = Union[SendCommandDevice, Union[float, int]]
SendCommandArg = Union[SendCommand, List[SendCommand]]

# Response from send commands.
SendCommandResponse = NamedTuple('SendCommandResponse',
                                 [('command', SendCommandDevice),
                                  ('code', str),
                                  ('msg', str)
                                  ])



# -*- coding: utf-8 -*-

"""
This only contains some helper routines that are used.
"""

import asyncio
import logging
from typing import List, Optional
from functools import partial

from aioharmony.handler import CallbackType

_LOGGER = logging.getLogger(__name__)


# pylint: disable=broad-except
def call_callback(callback_handler: CallbackType,
                  result: object,
                  callback_uuid: str,
                  callback_name: str) -> bool:
    # If we were provided a callback handler then call it now.
    if callback_handler is None:
        return False

    try:
        callback_result = call_raw_callback(
            callback=callback_handler,
            result=result,
            callback_uuid=callback_uuid,
            callback_name=callback_name
        )
    # Catching everything here.
    except Exception as exc:
        _LOGGER.exception("Exception in %s: %s",
                          callback_name,
                          exc)
        return False

    if not callback_result:
        _LOGGER.error("%s was not called due to mismatch in "
                      "callback type.",
                      callback_name
                      )
        return False

    return True


# TODO: Add this to Handler class
def call_raw_callback(callback: CallbackType,
                      result: object = None,
                      callback_uuid: str = None,
                      callback_name: str = None) -> bool:
    """
    Executes or sets the callback provided based on the type of callback:
      * Future : sets the result of the future to result provided
      * Event : sets the event
      * Coroutine: schedules the coroutine on the loop, result is
                   provided as the argument
      * Callback: Executes the callback, result is provided as the
                  argument

    :param callback: Handler that will be actioned
    :type callback: CallbackType
    :param callback_uuid: Unique number for the callback, only used in debug
                          logging
    :type callback_uuid:  str
    :param callback_name: Descriptive name for the callback, only used in
                          debug logging
    :type callback_name:  str
    :param result: object to set Future to, or to provide to the callback
                   routines
    :type result: object
    :return: True if callback was done successfully, False if it wasn't
    :rtype: bool
    """
    called_callback = False
    if asyncio.isfuture(callback):
        # It is a future, set the result of the future to
        # the message.
        if callback.done():
            _LOGGER.debug("Result of future %s with UUID %s was already "
                          "set",
                          callback_name,
                          callback_uuid)

        else:
            _LOGGER.debug("Future %s with UUID %s is set",
                          callback_name,
                          callback_uuid)
            callback.set_result(result)
            called_callback = True

        return called_callback

    if isinstance(callback, asyncio.Event):
        # It is an event, set the event.
        _LOGGER.debug("Setting event for handler %s with UUID %s",
                      callback_name,
                      callback_uuid)
        callback.set()
        return True

    if asyncio.iscoroutinefunction(callback):
        # Is a coroutine, schedule it on the loop.
        _LOGGER.debug("Scheduling coroutine %s with UUID %s",
                      callback_name,
                      callback_uuid)
        partial_func = partial(callback, result)
        partial_coro = asyncio.coroutine(partial_func)
        asyncio.ensure_future(partial_coro())
        return True

    if callable(callback):
        # This is a callback to be run. Execution of
        # these should be fast otherwise they're going
        # to block the loop.
        _LOGGER.debug("Calling callback %s with UUID %s",
                      callback_name,
                      callback_uuid)
        func = partial(callback, result)
        func()
        return True

    return False


def search_dict(match_value: object = None,
                key: str = None,
                search_list: List[dict] = None) -> Optional[dict]:
    """
    Returns the 1st element in a list containing dictionaries
    where the value of key provided matches the value provided.

    :param match_value: value to match upon (search for)
    :type match_value: object
    :param key: dictionary key to use for the match
    :type key: str
    :param search_list: List containing dictionary objects in which to search
    :type search_list: List[dict]
    :return: Dictionary object that matches
    :rtype: dict
    """
    if match_value is None or key is None or search_list is None:
        return None

    return next(
        (element for element in search_list
         if element[key] == match_value), None)



"""
Define the exception classes for aioharmony module.
"""


class HarmonyException(Exception):
    """Top level Harmony Exception"""


class HarmonyClient(HarmonyException):
    """
    Top level exception for HarmonyClient
    """


class TimeOut(HarmonyClient, TimeoutError):
    """
    Raised on timeouts
    """



# -*- coding: utf-8 -*-

"""

This is the module contains the classes for the callback handler. It is used
by :class:`~aioharmony.client.HarmonyClient` class to wait for responses from
the HUB through callbacks.

"""

import asyncio
import logging
from typing import (
    List, NamedTuple, Optional, Pattern, Union, Tuple
)
from uuid import uuid4
from datetime import datetime, timezone, timedelta
from aioharmony.helpers import call_callback
from aioharmony.handler import Handler

DEFAULT_TIMEOUT = 60

_LOGGER = logging.getLogger(__name__)

DataPatternType = Tuple[str, Pattern]

RespDataPatternType = Union[
    List[DataPatternType],
    DataPatternType
]

CallbackEntryType = NamedTuple('CallbackEntryType',
                               [('handler_uuid', str),
                                ('msgid', Optional[str]),
                                ('expiration', Optional[datetime]),
                                ('handler', Handler)
                                ])


class ResponseHandler:
    """

    Class to listen for json responses on the queue and then call
    registered handlers based on search patterns.

    This class is used by :class:`~aioharmony.client.HarmonyClient`, there is
    no need to use this class.

    The :class:`~aioharmony.client.HarmonyClient` class exposes methods
    :meth:`~ResponseHandler.register_handler` and
    :meth:`~ResponseHandler.unregister_handler` for registering
    additional handlers.

    :param message_queue: queue to listen on for JSON messages
    :type message_queue: asyncio.Queue
    """

    def __init__(self,
                 message_queue: asyncio.Queue) -> None:
        """"""
        self._message_queue = message_queue
        self._handler_list = []

        self._callback_task = asyncio.ensure_future(self._callback_handler())

    async def close(self):
        """Close all connections and tasks

           This should be called to ensure everything is stopped and
           cancelled out.
        """
        # Stop callback task
        if self._callback_task and not self._callback_task.done():
            self._callback_task.cancel()

    def register_handler(self,
                         handler: Handler,
                         msgid: str = None,
                         expiration: Union[
                             datetime,
                             timedelta] = None) -> str:
        """Register a handler.

        :param handler: Handler object to be registered
        :type handler: Handler
        :param msgid: Message ID to match upon.
                      DEFAULT = None
        :type msgid: Optional[str]
        :param expiration: How long or when handler should be removed. When
                           this is specified it will override what is set in
                           the Handler object.
                           If datetime is provided then UTC will be assumed
                           if tzinfo of the object is None.
                           DEFAULT = None
        :type expiration: Optional[Union[
                             datetime.datetime,
                             datetime.timedelta]]
        :return: Handler UUID number, this is a unique number for this handler
        :rtype: str
        """
        handler_uuid = str(uuid4())
        if expiration is None:
            expiration = handler.expiration

        if isinstance(expiration, timedelta):
            expiration = datetime.now(timezone.utc) + expiration

        if expiration is None:
            _LOGGER.debug("Registering handler %s with UUID %s",
                          handler.handler_name,
                          handler_uuid)
        else:
            if expiration.tzinfo is None:
                expiration = expiration.replace(tzinfo=timezone.utc)

            _LOGGER.debug("Registering handler %s with UUID %s that will "
                          "expire on %s",
                          handler.handler_name,
                          handler_uuid,
                          expiration.astimezone())

        self._handler_list.append(CallbackEntryType(
            handler_uuid=handler_uuid,
            msgid=msgid,
            expiration=expiration,
            handler=handler
        ))
        return handler_uuid

    def unregister_handler(self,
                           handler_uuid: str) -> bool:
        """Unregister a handler.

        :param handler_uuid: Handler UUID, this is returned by
                             register_handler when registering the handler
        :type handler_uuid: str
        :return: True if handler was found and thus deleted, False if it was
                 not found
        :rtype: bool
        """
        _LOGGER.debug("Unregistering handler with UUID %s",
                      handler_uuid)
        found_uuid = False
        for index in [index for index, element in enumerate(self._handler_list)
                      if element.handler_uuid == handler_uuid]:
            del self._handler_list[index]
            found_uuid = True
            break

        return found_uuid

    # pylint: disable=too-many-return-statements
    def _handler_match(self, dict_list, message, key=None):

        if key is not None:
            message = message.get(key)
            value = dict_list.get(key)
        else:
            value = dict_list

        if message is None or value is None:
            return False

        if isinstance(value, (dict, list)):
            # If they're different types then it is no match.
            # pylint: disable=unidiomatic-typecheck
            if type(message) != type(value):
                return False

            for new_key in value:
                if not self._handler_match(dict_list=value,
                                           message=message,
                                           key=new_key):
                    return False
            return True

        # value is a string or a pattern. If message is a dict or a list
        # then it is not a match.
        # Unable to check if message and value type are same when value is
        # not a list or dict as it can then be a string or a pattern whereas
        # message should be a string to do matching.
        if isinstance(message, (dict, list)):
            return False

        if isinstance(value, Pattern):
            if value.search(message) is None:
                return False
            return True

        return value == message

    def _get_handlers(self,
                      message: dict) -> List[CallbackEntryType]:
        """
        Find the handlers to be called for the JSON message received

        :param message: JSON message to use
        :type message: dict
        :return: List of Handler objects.
        :rtype: List[Handler]
        """

        callback_list = []
        for handler in self._handler_list:

            if handler.msgid is not None:
                if message.get('id') is None or \
                        message.get('id') != handler.msgid:
                    _LOGGER.debug("No match on msgid for %s",
                                  handler.handler.handler_name)
                    continue

            if handler.handler.resp_json is not None:
                if not self._handler_match(
                        dict_list=handler.handler.resp_json,
                        message=message):
                    _LOGGER.debug("No match for handler %s",
                                  handler.handler.handler_name)
                    continue

            _LOGGER.debug("Match for %s", handler.handler.handler_name)
            callback_list.append(handler)

        return callback_list

    def _unregister_expired_handlers(self,
                                     single_handler: CallbackEntryType =
                                     None) -> bool:
        """
        Unregisters any expired handlers based on their expiration
        datetime. Will check the handler dict instead of the list if provided
        :param single_handler:  Handler dict as it is put in the handler
                                list by register_handler.
                                DEFAULT = NONE
        :type single_handler: dict
        :return: True if one or more handlers were unregistered,
                 otherwise False
        :rtype: bool
        """

        if single_handler is None:
            handler_list = self._handler_list
        else:
            handler_list = [single_handler]

        removed_expired = False
        for handler in handler_list:
            if handler.expiration is not None:
                if datetime.now(timezone.utc) > handler.expiration:
                    _LOGGER.debug("Handler %s with UUID %s has "
                                  "expired, removing: %s",
                                  handler.handler.handler_name,
                                  handler.handler_uuid,
                                  handler.expiration.astimezone())
                    self.unregister_handler(handler.handler_uuid)
                    removed_expired = True

        return removed_expired

    # pylint: disable=broad-except
    async def _callback_handler(self) -> None:
        """
        Listens on the queue for JSON messages and then processes them by
        calling any handler(s)
        """
        _LOGGER.debug("Callback handler started")

        while True:
            # Put everything here in a try block, we do not want this
            # to stop running out due to an exception.
            try:
                # Wait for something to appear on the queue.
                message = await self._message_queue.get()
                _LOGGER.debug("Message received: %s", message)

                # Go through list and call
                for handler in self._get_handlers(message=message):

                    # Make sure handler hasn't expired yet.
                    if self._unregister_expired_handlers(
                            single_handler=handler):
                        # Was expired and now removed, go on with next one.
                        continue

                    call_callback(
                        callback_handler=handler.handler.handler_obj,
                        result=message,
                        callback_uuid=handler.handler_uuid,
                        callback_name=handler.handler.handler_name
                    )

                    # Remove the handler from the list if it was only to be
                    # called once.
                    if handler.handler.once:
                        self.unregister_handler(handler.handler_uuid)

                # Go through all handlers and remove expired ones IF
                # currently
                # nothing in the queue.
                if self._message_queue.empty():
                    # Go through list and remove all expired ones.
                    _LOGGER.debug("Checking for expired handlers")
                    self._unregister_expired_handlers()

            except asyncio.CancelledError:
                _LOGGER.debug("Received STOP for callback handler")
                break

            # Need to catch everything here to prevent an issue in a
            # from causing the handler to exit.
            except Exception as exc:
                _LOGGER.exception("Exception in callback handler: %s", exc)

        # Reset the queue.
        self._message_queue = asyncio.Queue()

        _LOGGER.debug("Callback handler stopped.")



# -*- coding: utf-8 -*-

"""
This is the main module containing the class to be imported and used:
from aioharmony.harmonyapi import HarmonyAPI

The HarmonyAPI class is a wrapper around the HarmonyClient class
which represents the Harmony Hub.
Using the methods of this class allows one to query or send commands to the
Hub.
"""

import asyncio
import json
import logging
from datetime import datetime, timedelta
from typing import List, Optional, Union
from aioharmony.const import (
    ClientConfigType, SendCommandArg, SendCommandDevice, SendCommandResponse
)
from aioharmony.harmonyclient import ClientCallbackType, HarmonyClient
from aioharmony.handler import Handler

_LOGGER = logging.getLogger(__name__)

# Making these types available for import.
ClientConfigType = ClientConfigType
SendCommandDevice = SendCommandDevice

# TODO: Add docstyle comments
# TODO: Clean up code styling


# pylint: disable=too-many-public-methods
class HarmonyAPI:

    # pylint: disable=too-many-arguments
    def __init__(self,
                 ip_address: str,
                 callbacks: ClientCallbackType = None,
                 loop: asyncio.AbstractEventLoop = None) -> None:
        _LOGGER.debug("%s: Initialize", ip_address)
        loop = loop if loop else asyncio.get_event_loop()

        self._harmony_client = HarmonyClient(
            ip_address=ip_address,
            callbacks=callbacks,
            loop=loop
        )

    @property
    def ip_address(self) -> str:
        return self._harmony_client.ip_address

    @property
    def hub_config(self) -> ClientConfigType:
        return self._harmony_client.hub_config

    @property
    def name(self) -> Optional[str]:
        return self._harmony_client.name

    @property
    def email(self) -> Optional[str]:
        return self.hub_config.info.get('email')

    @property
    def account_id(self) -> Optional[str]:
        return self.hub_config.info.get('accountId')

    @property
    def fw_version(self) -> Optional[str]:
        return self.hub_config.info.get('current_fw_version')

    @property
    def current_activity(self) -> tuple:
        return self._harmony_client.current_activity_id, \
            self._harmony_client.get_activity_name(
                self._harmony_client.current_activity_id)

    @property
    def config(self) -> dict:
        return self.hub_config.config

    @property
    def json_config(self) -> dict:
        """Returns configuration as a dictionary (json)"""

        result = {}
        config = self.config
        activity_dict = {}

        for activity in config.get('activity', []):
            activity_dict.update({activity['id']: activity['label']})

        result.update(Activities=activity_dict)

        devices_dict = {}
        for device in config.get('device', []):
            command_list = []
            for control_group in device.get('controlGroup', []):
                for function in control_group.get('function', []):
                    action = json.loads(function.get('action'))
                    if action is not None:
                        command_list.append(action.get('command'))

            device_dict = {
                'id': device.get('id'),
                'commands': command_list
            }

            devices_dict.update({device.get('label'): device_dict})

        result.update(Devices=devices_dict)

        return result

    @property
    def callbacks(self) -> ClientCallbackType:
        return self._harmony_client.callbacks

    @callbacks.setter
    def callbacks(self, value: ClientCallbackType) -> None:
        self._harmony_client.callbacks = value

    def get_activity_id(self, activity_name) -> Optional[str]:
        return self._harmony_client.get_activity_id(
            activity_name=activity_name)

    def get_activity_name(self, activity_id) -> Optional[str]:
        return self._harmony_client.get_activity_name(activity_id=activity_id)

    def get_device_id(self, device_name) -> Optional[str]:
        return self._harmony_client.get_device_id(device_name=device_name)

    def get_device_name(self, device_id) -> Optional[str]:
        return self._harmony_client.get_device_name(device_id=device_id)

    async def connect(self) -> bool:
        return await self._harmony_client.connect()

    async def close(self) -> None:
        await self._harmony_client.close()

    def register_handler(self,
                         handler: Handler,
                         msgid: str = None,
                         expiration: Union[
                             datetime,
                             timedelta] = None) -> str:
        """Register a handler.

        :param handler: Handler object to be registered
        :type handler: Handler
        :param msgid: Message ID to match upon.
                      DEFAULT = None
        :type msgid: Optional[str]
        :param expiration: How long or when handler should be removed. When
                           this is specified it will override what is set in
                           the Handler object.
                           If datetime is provided then UTC will be assumed
                           if tzinfo of the object is None.
                           DEFAULT = None
        :type expiration: Optional[Union[
                             datetime.datetime,
                             datetime.timedelta]]
        :return: Handler UUID number, this is a unique number for this handler
        :rtype: str
        """
        return self._harmony_client.register_handler(
            handler=handler,
            msgid=msgid,
            expiration=expiration)

    def unregister_handler(self,
                           handler_uuid: str) -> bool:
        """Unregister a handler.

        :param handler_uuid: Handler UUID, this is returned by
                             register_handler when registering the handler
        :type handler_uuid: str
        :return: True if handler was found and thus deleted, False if it was
                 not found
        :rtype: bool
        """
        return self._harmony_client.unregister_handler(
            handler_uuid=handler_uuid)

    async def sync(self) -> bool:
        """Syncs the harmony hub with the web service."""
        _LOGGER.debug("%s: Performing sync", self.name)
        # Send the command to the HUB
        response = await self._harmony_client.send_to_hub(command='sync')
        if not response or response.get('code') != 200:
            # There was an issue
            return False

        # Update our own information.
        await self._harmony_client.refresh_info_from_hub()
        return True

    async def start_activity(self, activity_id) -> tuple:
        return await self._harmony_client.start_activity(
            activity_id=activity_id)

    async def send_commands(self,
                            commands: SendCommandArg) -> \
            List[SendCommandResponse]:

        if isinstance(commands, list):
            _LOGGER.debug("%s: Sending commands to HUB", self.name)
        else:
            _LOGGER.debug("%s: Sending command to HUB", self.name)
            # Changing it to list.
            commands = [commands]

        return await self._harmony_client.send_commands(commands=commands)

    async def power_off(self) -> bool:
        """Turns the system off if it's on, otherwise it does nothing.

        Returns:
            True if the system becomes or is off
        """
        result = await self.start_activity(-1)
        return result[0]

    async def change_channel(self, channel: int) -> bool:
        """Change channel

        :param channel: Channel number
        :type channel: int
        :return: True if successfully, False if unsuccessfully
        :rtype: bool
        """
        _LOGGER.debug("%s: Changing channel to %s",
                      self.name,
                      channel)
        params = {
            "timestamp": 0,
            'channel': str(channel)
        }

        # Send the command to the HUB
        response = await self._harmony_client.send_to_hub(
            command='change_channel',
            params=params)
        if not response:
            # There was an issue
            return False

        return response.get('code') == 200
